package cz.cvut.fel.pr1.airports;

import com.opencsv.CSVReader;
import com.opencsv.CSVReaderBuilder;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;

public class Airports {

    public static final double EARTH_RADIUS_KM = 6372.8;

    public static final int COL_ID = 1;
    public static final int COL_NAME = 3;
    public static final int COL_LATITUDE = 4;
    public static final int COL_LONGITUDE = 5;

    private String[] names = new String[0];
    private String[] ids = new String[0];
    private double[] latitudes = new double[0];
    private double[] longitudes = new double[0];

    private final boolean useBinarySearch;

    /**
     * Creates new instance of Airports handler
     * @param useBinarySearch whether to use (faster) binary search
     */
    public Airports(boolean useBinarySearch) {
        this.useBinarySearch = useBinarySearch;
    }


    /**
     * Loads data from given {@code filename} and prepares them for work.
     * @param filename from which are data loaded
     */
    public void loadData(String filename) {
        final Path path = Paths.get(filename);
        List<String[]> data = loadFile(path);
        assert data != null;

        parseData(data);

        if (useBinarySearch) {
            sort();
        }
    }

    /**
     * Calculates distance between given airports and returnes the value. If the {@code print} option is {@code
     * true}, the informative line is printed to stdout.
     * @param fromId id of the first airport
     * @param toId id of the second airport
     * @param print wheter to print out info
     * @return distance between given airports in km
     */
    public double distanceBetweenAirports(String fromId, String toId, boolean print) {
        final int fromIndex= indexOfAirport(fromId);
        final int toIndex = indexOfAirport(toId);

        final double fromLatitude = latitudes[fromIndex];
        final double fromLongitude = longitudes[fromIndex];

        final double toLatitude = latitudes[toIndex];
        final double toLongitude = longitudes[toIndex];

        final double distanceInKm = calculateDistance(fromLatitude, fromLongitude, toLatitude, toLongitude);

        if (print) {
            final String fromName = names[fromIndex];
            final String toName = names[toIndex];
            System.out.printf("Distance between %s-%s and %s-%s is %f km%n", fromId, fromName, toId, toName, distanceInKm);
        }

        return distanceInKm;
    }

    public double latitudeOfAirport(String airportId) {
        final int airportIndex = indexOfAirport(airportId);
        return latitudes[airportIndex];
    }

    public double longitudeOfAirport(String airportId) {
        final int airportIndex = indexOfAirport(airportId);
        return longitudes[airportIndex];
    }

    public String nameOfAirport(String airportId) {
        final int airportIndex = indexOfAirport(airportId);
        return names[airportIndex];
    }

    private int indexOfAirport(String airportId) {
        if (useBinarySearch) {
            return indexOfAirportBinary(airportId);
        } else {
            return indexOfAirportLinear(airportId);
        }
    }

    /**
     * Finds the index of airport with given ID using linear search.
     * @param airportId id of airport
     * @return index of the id or -1 if the ID was not found
     */
    private int indexOfAirportLinear(String airportId) {
        int out = -1;
        for (int i = 0; i < ids.length; i++) {
            if (ids[i].equals(airportId)) {
                out = i;
                break;
            }
        }
        return out;
    }

    /**
     * Finds the index of airport with given ID using binary search. Note that the {@link #ids} array must be sorted
     * alphabetically beforehand.
     * @param airportId id of airport
     * @return index of the id or -1 if the ID was not found
     */
    private int indexOfAirportBinary(String airportId) {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /**
     * Calculates distance between gps coordinates defined by ({@code fromLatitude}, {@code fromLongitude}) and
     * ({@code toLatitude}, {@code toLongitude}). The algorithm uses haversine distance approximation. Code was
     * taken from <a href="http://rosettacode.org/wiki/Haversine_formula#Java">rosettacode website</a>.
     *
     * @param fromLatitude latitude of the first point
     * @param fromLongitude longitude of the first point
     * @param toLatitude latitude of the second point
     * @param toLongitude longitude of the second point
     * @return distance in km
     */
    private double calculateDistance(double fromLatitude, double fromLongitude, double toLatitude, double toLongitude) {
        double dLat = Math.toRadians(toLatitude - fromLatitude);
        double dLon = Math.toRadians(toLongitude - fromLongitude);
        fromLatitude = Math.toRadians(fromLatitude);
        toLatitude = Math.toRadians(toLatitude);

        double a = Math.pow(Math.sin(dLat / 2),2) + Math.pow(Math.sin(dLon / 2),2) * Math.cos(fromLatitude) * Math.cos(toLatitude);
        double c = 2 * Math.asin(Math.sqrt(a));
        return EARTH_RADIUS_KM * c;
    }

    private List<String[]> loadFile(Path path) {
        if (!Files.isReadable(path)) {
            System.err.println("Cound not find given file "+ path);
            System.err.println("Did you download the file using command " +
                    "'curl https://cw.fel.cvut.cz/wiki/_media/courses/a0b36pr1/labs/airports.tar.gz | tar zxf -'");
            System.exit(1);
        }

        List<String[]> data = null;

        try (BufferedReader bufferedReader = Files.newBufferedReader(path);
             CSVReader csvReader = new CSVReaderBuilder(bufferedReader).withSkipLines(1).build()) {
            data = csvReader.readAll();
        } catch (IOException e) {
            System.err.println("Error occured while loading "+ path);
            e.printStackTrace();
            System.exit(2);
        }
        return data;
    }

    private void parseData(List<String[]> data) {
        initArrays(data.size());

        int index = 0;
        for (String[] line : data) {
            if (!line[COL_ID].isEmpty()) {
                ids[index] = line[COL_ID].toUpperCase();
                names[index] = line[COL_NAME];
                latitudes[index] = Double.parseDouble(line[COL_LATITUDE]);
                longitudes[index] = Double.parseDouble(line[COL_LONGITUDE]);
                index++;
            }
        }

        trimArray(index);
    }

    private void initArrays(int size) {
        ids = new String[size];
        names = new String[size];
        latitudes = new double[size];
        longitudes = new double[size];
    }

    private void trimArray(int index) {
        ids = Arrays.copyOf(ids, index);
        names = Arrays.copyOf(names, index);
        latitudes = Arrays.copyOf(latitudes, index);
        longitudes = Arrays.copyOf(longitudes, index);
    }

    /**
     * Sorts loaded data alphabetically by ids.
     */
    private void sort() {
        throw new UnsupportedOperationException("Not yet implemented");
    }


    public static void main(String[] args) {
        final boolean useBinary = false;
        final boolean print = true;

        final Airports airports = new Airports(useBinary);
        airports.loadData("airports.csv");

        final long startTimeStamp = System.currentTimeMillis();
        for (int i = 0; i < 1000; i++) {
            airports.distanceBetweenAirports(Identificators.PRAGUE, Identificators.LONDON, print);
            airports.distanceBetweenAirports(Identificators.HONG_KONG, Identificators.MELBOURNE, print);
            airports.distanceBetweenAirports(Identificators.LOS_ANGELES, Identificators.TOKYO, print);
            airports.distanceBetweenAirports(Identificators.NAYPYIDAW, Identificators.ANTANANARIVO, print);
            airports.distanceBetweenAirports(Identificators.BRUSSELS, Identificators.MONTREAL, print);
            airports.distanceBetweenAirports(Identificators.KAMPALA, Identificators.PRAGUE, print);
        }
        final long stopTimeStamp = System.currentTimeMillis();

        System.out.printf("The batch took %d ms%n", stopTimeStamp - startTimeStamp);

        //new KmlGenerator(airports).generateKml();
    }
}
